#include "rusinluk_dip_incl.h"

//#include "rusinluk_dip_test.hpp"

void resynthTask(const RWork &work, const Resynth &resynth, const Params &pm, SharedMem &mem) {
	string file = pm.fileI;
	fileSuffix(file, FILE_CWIN + toString(work.cwinI) + FILE_METHOD + toString(work.methodI));
	resynth.run(work.methodI, file, pm, mem);
}

int chooseBest0(const string &file, int methodN, const Params &pm) {
	if (!methodN) {
		return NO_INDEX;
	}
	int res = NO_INDEX;
	fileTest(file); // for an unexpected case
	Circuit circ(file);
	double
		val = circ.eval(pm),
		minVal = val;
	string fileW;
	for (int i = 0; i < methodN; i++) {
		fileW = file;
		fileSuffix(fileW, FILE_METHOD + toString(i));
		fileTest(fileW);
		circ.load(fileW);
		if ((val = circ.eval(pm)) < minVal) {
			res = i;
			minVal = val;
		}
	}
	return res;
}

/* renames the best resynth output (or input, in case of choice as best)
*/
void chooseBest(int cwinI, int methodN, const Params &pm) {
	string
		fileI = pm.fileI,
		fileO = pm.fileI;
	fileSuffix(fileI, FILE_CWIN + toString(cwinI));
	int res = chooseBest0(fileI, methodN, pm);
	if (res != NO_INDEX) {
		fileSuffix(fileI, FILE_METHOD + toString(res));
	}
	fileSuffix(fileO, FILE_OUTPUT + toString(cwinI));
	if (rename(fileI.c_str(), fileO.c_str()) != 0) {
		logErr("Error renaming file " + fileI + " to " + fileO);
	}
}

void clear(int cwinN, int methodN, const Params &pm) {
	string file, file2;
	for (int i = 0; i < cwinN; i++) {
		file = pm.fileI;
		fileSuffix(file, FILE_CWIN + toString(i));
		remove(file.c_str());
		for (int j = 0; j < methodN; j++) {
			file2 = file;
			fileSuffix(file2, FILE_METHOD + toString(j));
			remove(file2.c_str());
		}
		file = pm.fileI;
		fileSuffix(file, FILE_OUTPUT + toString(i));
		remove(file.c_str());
	}
}

/* each resynth method will have its own copy of input file
- to avoid collective reading collisions in external scripts
*/
void copyForResynth(int cwinN, int methodN, const Params &pm) {
	string file, file2;
	for (int i = 0; i < cwinN; i++) {
		file = pm.fileI;
		fileSuffix(file, FILE_CWIN + toString(i));
		for (int j = 0; j < methodN; j++) {
			file2 = file;
			fileSuffix(file2, FILE_METHOD + toString(j));
			fileCopy(file, file2);
		}
	}
}

void resynthPart(int cwinN, const Resynth &resynth, const Params &pm, double time0) {
	int
		rank = mpiRank(),
		procN = mpiProcN();
	Mutex mutex(TAG_MUTEX, pm.log);
	Scheduler schd(cwinN, resynth.getMethodN());
	SharedMem mem(schd.getBufN());
	if (rank == BOSS_RANK) {
		schd.rmaPut(mem);
	}
	mpiBarrier();
	if (rank != BOSS_RANK) {
		schd.rmaGet(mem);
	}
	RWork work;
	for (int i = procN - 1; i >= rank; i--) {
		work.setInvalid();
		schd.submitWork(work); // aka get next, initial distribution
	}
	mpiBarrier();
	if (rank == 0) {
		schd.rmaPut(mem);
	}
	mpiBarrier();
	if (pm.log) { logTime("Resynthesis - begin.", time0); }
	while (work.isValid()) {
		if (work.isEval()) {
			chooseBest(work.cwinI, resynth.getMethodN(), pm);
		}
		else {
			resynthTask(work, resynth, pm, mem);
		}
		if (pm.log) { logTime("Resynthesis task " + work.getStr() + " done.", time0); }
		mutex.lock();
		{
			if (pm.log) { logTime("Resynthesis - locked.", time0); }
			schd.rmaGet(mem);
			schd.submitWork(work);
			if (pm.log) { logRow("Work submitted. New work: " + work.getStr()); }
			schd.rmaPut(mem);
		}
		mutex.unlock();
		if (pm.log) { logTime("Resynthesis - unlocked.", time0); }
	}
	if (pm.log) { logTime("Waiting for others...", time0); }
	mpiBarrier();
	if (rank == BOSS_RANK) {
		schd.rmaGet(mem);
		if (!schd.isDone()) {
			logErr("Scheduler error - undone.");
		}
	}
		// blocking in mems' destructors
}

void init(int argc, char *argv[], Params &pm, MyRand &rnd) {
	int rank = mpiRank();
	if (!pm.load(argc, argv)) {
		if (rank == BOSS_RANK) {
			pm.printArgs();
			mpiAbort();
		}
		mpiBarrier();
	}
	if (rank == BOSS_RANK) {
		if (pm.methods.size()) {
			fileTest(pm.methods);
		}
	}
	if (!pm.rand) {
		rnd.seed(pm.rseed1, pm.rseed2);
	}
}

int main(int argc, char *argv[]) {
	MPI_Init(&argc, &argv);
	double time0 = MPI_Wtime();
	int
		procN = mpiProcN(),
		rank = mpiRank();
	Params pm;
	MyRand rnd;
	init(argc, argv, pm, rnd);
	if (rank == BOSS_RANK) {
		logRow("Number of processes = " + toString(procN));
		pm.print();
	}
	waitPhase(WAIT_PHASE);
	Resynth resynth(pm.methods);
	Circuit *circ = NULL;
	int cwinN;
	if (rank == BOSS_RANK) {
		if (pm.log) { logRow("Loading the circuit..."); }
		circ = new Circuit(pm.fileI);
		if (pm.log) { logTime("Circuit loaded.", time0); }
		circ->printEval();
		if (pm.log) { logRow("Making circuit windows..."); }
		circ->makeWins(pm, rnd);
		cwinN = circ->getCwinN();
		logTime(toString(cwinN) + " windows made in the circuit.", time0);
		if (pm.log) { logRow("Splitting the circuit..."); }
		circ->split(pm);
		if (pm.log) { logTime("Circuit splitted.", time0); }
		copyForResynth(circ->getCwinN(), resynth.getMethodN(), pm);
		if (pm.log) { logTime("Circuit windows copied for resynthesis.", time0); }
	}
	else {
		if (pm.log) { logRow("Waiting for the boss..."); }
	}
	mpiBarrier();
	if (pm.log) { logRow("Starting the resynthesis..."); }
	resynthPart((circ != NULL ? circ->getCwinN() : 0), resynth, pm, time0);
	if (pm.log) { logTime("Resynthesis part done.", time0); }
	if (rank == BOSS_RANK) {
		if (pm.log) { logRow("Merging the circuit..."); }
		circ->mergeAndSave(pm.fileO);
		if (pm.log) { logTime("Circuit merged.", time0); }
		if (pm.clear) {
			if (pm.log) { logRow("Removing working files..."); }
			clear(circ->getCwinN(), resynth.getMethodN(), pm);
			if (pm.log) { logTime("Files removed.", time0); }
		}
		if (pm.log) { logRow("Evaluating the result..."); }
		circ->load(pm.fileO);
		circ->printEval();
		logRow(
			"| " + toString(procN) +
			" | " + toString(cwinN) +
			" | " + toString(circ->getGateN()) +
			" | " + toString(circ->getGateInN()) +
			" | " + toString(circ->getLevN()) +
			" | " + toString(floor(MPI_Wtime() - time0 + 0.5)) + " |"
		);
	}
	if (pm.log || rank == BOSS_RANK) {
		logTime("Program done.", time0);
		logRow("");
	}
	delete circ;
	MPI_Finalize();
	return 0;
}
