#include "rusinluk_dip_incl.h"

Mutex::Mutex(int tag, bool log) :
	size(mpiProcN() + 1),
	mem(size),
	data(new char[size]),
	lockI(size - 1),
	locked(false),
	tag(tag),
	log(log)
{
	if (mpiRank() == BOSS_RANK) {
		for (int i = 0; i < size; i++) {
			data[i] = 0;
		}
		putData();
	}
	mpiBarrier();
}

Mutex::~Mutex() {
	delete[] data;
}

bool Mutex::lock() {
	if (locked) {
		return false;
	}
	int
		myRank = mpiRank(),
		procN = mpiProcN();
	int val, src, rank; // message, its source, probe rank
	put(true, myRank);
	if (log) { logRow("Mutex: Arrived."); }
		// everyone knows that i'm able to communicate from now
	while (1) {
		while (get(lockI)) {
			recv(val, src);
			switch (val) {
				case MUTEX_POKE :
					if (log) { logRow("Mutex: Poked."); }
					break; // => the next lock check
				case MUTEX_PROBE :
				case MUTEX_WINNER :
					send(MUTEX_PASSIVE, src); // I can't win from here
					break;
				case MUTEX_TICKET :
					if (log) { logRow("Mutex: Ticket has been used."); }
					return (locked = true); // I've got a ticket!
				default :
					logErr(ERR_MUTEX, "in waiting loop");
			}
		}
		put(true, lockI);
		if (log) { logRow("Mutex: Locked."); }
			/* locked, no one new from outside can access this point now
			=> closed group of processes here, the lowest rank will pass the lock */
		getData();
		int
			activeI = NO_INDEX, // message from neighbor on the left (his rank)
			winnerI = NO_INDEX, // if my left neighbor is the winner
			probeI = NO_INDEX; // probe from neighbor on the right/winner
		bool winner = true; // am I the winner?
		for (rank = myRank - 1; rank >= 0; rank--) {
			if (!data[rank]) {
				continue;
			}
			send(MUTEX_PROBE, rank);
			while (1) {
				recv2(val, src);
				if (val == MUTEX_PASSIVE) {
					break;
				}
				if (val == MUTEX_ACTIVE) {
					activeI = src;
					break;
				}
				if (val == MUTEX_WINNER) {
					winnerI = src;
					if (winnerI == rank) {
						break;
					}
					continue;
				}
				if (val == MUTEX_PROBE) {
					probeI = src;
					continue;
				}
				logErr(ERR_MUTEX, "in locking loop #1");
			}
			if (winnerI != NO_INDEX && winnerI != rank) {
				send(MUTEX_PROBE, winnerI);
			}
			if (activeI != NO_INDEX || winnerI != NO_INDEX) {
				winner = false; // I'm not the most left active
				break;
			}
		}
		if (!winner) {
			if (probeI == NO_INDEX) {
				if (log) { logRow("Mutex: Waiting for a probe..."); }
				recv2(val, src);
				if (val != MUTEX_PROBE) {
					logErr(ERR_MUTEX, "No probe received.");
				}
				probeI = src;
			}
			send(MUTEX_ACTIVE, probeI); // breaking the chain of dependency
			if (log) { logRow("Mutex: Go passive."); }
			continue; // back to waiting loop
		}
			// I'm the winner!
		if (probeI != NO_INDEX) {
			send(MUTEX_WINNER, probeI);
		}
		else {
			for (rank = myRank + 1; rank < procN; rank++) {
				if (!data[rank]) {
					continue;
				}
				send(MUTEX_WINNER, rank);
				while (1) {
					recv2(val, src);
					if (val == MUTEX_PASSIVE) {
						break;
					}
					if (val == MUTEX_PROBE) {
						probeI = src;
						if (probeI == rank) {
							break;
						}
						continue;
					}
					logErr(ERR_MUTEX, "in locking loop #2");
				}
				if (probeI != NO_INDEX) {
					if (probeI != rank) {
						send(MUTEX_WINNER, probeI); /* reply to neighbor's probe only in this case,
							otherwise the "reply" has been already sent! */
					}
					break;
				}
			}
			if (probeI == NO_INDEX) {
				break; // alone, I can pass
			}
		}
		activeI = NO_INDEX;
		for (rank = procN - 1; rank > myRank; rank--) {
			if (!data[rank]) {
				continue;
			}
			send(MUTEX_PROBE, rank);
			while (1) {
				recv2(val, src);
				if (val == MUTEX_PASSIVE) {
					break;
				}
				if (val == MUTEX_ACTIVE) {
						// the last message has been received
					activeI = src;
					break;
				}
				logErr(ERR_MUTEX, "in locking loop #3");
			}
			if (activeI != NO_INDEX) {
				break;
			}
		}
		if (activeI == NO_INDEX) {
			logErr(ERR_MUTEX, "No ack to pass the lock.");
		}
		break; // no communication in progress => I can pass
	}
	if (log) { logRow("Mutex: Passed the lock."); }
	return (locked = true);
}

bool Mutex::unlock() {
	if (!locked) {
		return false;
	}
	int
		rank = mpiRank(),
		procN = mpiProcN();
	put(false, rank); // leaving
	int
		pos = rnd.next() % procN,
		ticketI = NO_INDEX;
	getData();
	for (int i = 0; i < procN; i++) {
		int j = (pos + i) % procN;
		if (data[j]) {
			ticketI = j;
			break;
		}
	}
	if (ticketI != NO_INDEX) {
		send(MUTEX_TICKET, ticketI); // there's someone to give him a ticket (optimization)
		if (log) { logRow("Mutex: A ticket has been sent to: " + toString(ticketI)); }
	}
	else {
		put(false, lockI); // I have to unlock
		if (log) { logRow("Mutex: Unlocked."); }
			// deadlock prevency
		getData();
		pos = rnd.next() % procN;
		for (int i = 0; i < procN; i++) {
			int j = (pos + i) % procN;
			if (data[j]) {
				send(MUTEX_POKE, j); // => at least 1 process will pass the lock
				if (log) { logRow("Mutex: Process " + toString(j) + " has been poked."); }
				break;
			}
		}
	}
	locked = false;
	return true;
}

bool Mutex::get(int adr) {
	char res;
	mem.lock();
	int a = MPI_Get(&res, 1, MPI_BYTE, BOSS_RANK, adr, 1, MPI_BYTE, mem.win);
	if (a != MPI_SUCCESS) {
		logErr("Mutex: RMA receiving error (" + toString(a) + ")");
	}
	mem.unlock();
	return (res != 0);
}

void Mutex::put(bool val0, int adr) {
	char val = (char)val0;
	mem.lock();
	int a = MPI_Put(&val, 1, MPI_BYTE, BOSS_RANK, adr, 1, MPI_BYTE, mem.win);
	if (a != MPI_SUCCESS) {
		logErr("Mutex: RMA sending error (" + toString(a) + ")");
	}
	mem.unlock();
}

void Mutex::getData() {
	mem.lock();
	int a = MPI_Get(data, size, MPI_BYTE, BOSS_RANK, 0, size, MPI_BYTE, mem.win);
	if (a != MPI_SUCCESS) {
		logErr("Mutex: RMA receiving error (data, " + toString(a) + ")");
	}
	mem.unlock();
}

void Mutex::putData() {
	mem.lock();
	int a = MPI_Put(data, size, MPI_BYTE, BOSS_RANK, 0, size, MPI_BYTE, mem.win);
	if (a != MPI_SUCCESS) {
		logErr("Mutex: RMA sending error (data, " + toString(a) + ")");
	}
	mem.unlock();
}

void Mutex::send(int val, int dest) const {
	MPI_Request req;
	int a = MPI_Isend(&val, 1, MPI_INT, dest, tag, MPI_COMM_WORLD, &req);
	if (a != MPI_SUCCESS) {
		logErr("Mutex: Message sending error (" + toString(val) + ", " + toString(dest) + ", " + toString(a) + ")");
	}
	int flag = 0;
	MPI_Status status;
	do {
		MPI_Test(&req, &flag, &status);
	}
		while (!flag);
	if (log) { logRow("Mutex: Sent message " + toString(val) + " to process " + toString(dest)); }
}

void Mutex::recv(int &val, int &src) const {
	MPI_Status status;
	int a = MPI_Recv(&val, 1, MPI_INT, MPI_ANY_SOURCE, tag, MPI_COMM_WORLD, &status);
	if (a != MPI_SUCCESS) {
		logErr("Mutex: Message receiving error (" + toString(val) + ", " + toString(a) + ")");
	}
	src = status.MPI_SOURCE;
	if (log) { logRow("Mutex: Received message " + toString(val) + " from process " + toString(src)); }
}

void Mutex::recv2(int &val, int &src) const {
	while (1) {
		recv(val, src);
		if (val != MUTEX_POKE) {
			break;
		}
		if (log) { logRow("Mutex: Poke ignored."); }
	}
}
